from enum import Enum


class Accelerator(str, Enum):
    nvidia_t4 = "nvidia_t4"
    nvidia_a100 = "nvidia_a100"
    nvidia_a100_80gb = "nvidia_a100_80gb"
    nvidia_h100 = "nvidia_h100"
    nvidia_l4 = "nvidia_l4"

    cpu = "cpu"

    nvidia_a100_5gb = "nvidia_a100_5gb"
    nvidia_a100_10gb = "nvidia_a100_10gb"
    nvidia_a100_20gb = "nvidia_a100_20gb"

    nvidia_a100_80gb_10gb = "nvidia_a100_80gb_10gb"
    nvidia_a100_80gb_20gb = "nvidia_a100_80gb_20gb"
    nvidia_a100_80gb_40gb = "nvidia_a100_80gb_40gb"

    nvidia_a10 = "nvidia_a10"
    nvidia_a10_12gb = "nvidia_a10_12gb"
    nvidia_a10_8gb = "nvidia_a10_8gb"
    nvidia_a10_4gb = "nvidia_a10_4gb"

    nvidia_h100_10gb = "nvidia_h100_10gb"
    nvidia_h100_20gb = "nvidia_h100_20gb"
    nvidia_h100_40gb = "nvidia_h100_40gb"
    nvidia_h100_80gb = "nvidia_h100_80gb"



from .build import build_container
from .init_dir import init_dir
from .push import push_container
from .up import up_container

__all__ = [
    "build_container",
    "init_dir",
    "push_container",
    "up_container",
]



import json
import os
import sys
from argparse import Namespace
from pathlib import Path

import docker
import docker.errors
import yaml

from pipeline.cloud import http
from pipeline.cloud.schemas import cluster as cluster_schemas
from pipeline.cloud.schemas import pipelines as pipelines_schemas
from pipeline.cloud.schemas import registry as registry_schemas
from pipeline.util.logging import _print

from .pointer import create_pointer
from .schemas import PipelineConfig


def push_container(namespace: Namespace):
    """
    Upload protocol:
    1. Request upload URL from server, along with auth token
    2. Upload to URL with auth token
    3. Send complete request to server
    """

    config_file = Path(getattr(namespace, "file", "./pipeline.yaml"))

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pointers: list | None = getattr(namespace, "pointer", None)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    # parse optional cluster CLI arg
    cluster_id: str | None = getattr(namespace, "cluster", None)
    node_pool_id: str | None = getattr(namespace, "node_pool", None)
    # Ensure node pool arg is provided if cluster arg provided
    if cluster_id and node_pool_id is None:
        raise ValueError("If --cluster is provided, --node-pool must also be provided")
    if node_pool_id and cluster_id is None:
        raise ValueError("If --node-pool is provided, --cluster must also be provided")
    if cluster_id and node_pool_id:
        pipeline_config.cluster = cluster_schemas.PipelineClusterConfig(
            id=cluster_id, node_pool=node_pool_id
        )

    # Check for file, transform to string, and put it back in config
    if pipeline_config.readme is not None:
        if os.path.isfile(pipeline_config.readme):
            markdown_file = Path(pipeline_config.readme)
            pipeline_config.readme = markdown_file.read_text()
    else:
        pipeline_config.readme = ""

    # Attempt to format the readme
    readmeless_config = pipeline_config.copy()
    readmeless_config.readme = None
    pipeline_yaml_text = yaml.dump(
        json.loads(readmeless_config.json(exclude_none=True, exclude_unset=True)),
        sort_keys=False,
    )

    pipeline_yaml_text = "```yaml\n" + pipeline_yaml_text + "\n```"
    pipeline_code = Path(
        pipeline_config.pipeline_graph.split(":")[0] + ".py"
    ).read_text()
    pipeline_code = "```python\n" + pipeline_code + "\n```"
    pipeline_config.readme = pipeline_config.readme.format(
        pipeline_name=pipeline_config.pipeline_name,
        pipeline_yaml=pipeline_yaml_text,
        pipeline_code=pipeline_code,
    )

    pipeline_name = (
        pipeline_config.pipeline_name.split(":")[0]
        if ":" in pipeline_config.pipeline_name
        else pipeline_config.pipeline_name
    )

    docker_client: docker.DockerClient = docker.from_env(timeout=600)

    registry_params = {}
    if pipeline_config.extras and pipeline_config.extras.get("turbo_registry", False):
        registry_params["turbo_registry"] = "yes"

    registry_info = http.get(endpoint="/v4/registry", params=registry_params)
    registry_info = registry_schemas.RegistryInformation.parse_raw(registry_info.text)

    upload_registry = registry_info.url

    if upload_registry is None:
        raise ValueError("No upload registry found")
    image = docker_client.images.get(pipeline_name)
    image_hash = image.id.split(":")[1]

    hash_tag = image_hash[:12]
    image_to_push = pipeline_name + ":" + hash_tag
    image_to_push_reg = upload_registry + "/" + image_to_push

    upload_token = None
    true_pipeline_name = None
    if registry_info.special_auth:
        start_upload_response = http.post(
            endpoint="/v4/registry/start-upload",
            json_data=pipelines_schemas.PipelineStartUpload(
                pipeline_name=pipeline_name,
                pipeline_tag=None,
                cluster=pipeline_config.cluster,
            ).dict(),
        )
        start_upload_dict = start_upload_response.json()
        upload_token = start_upload_dict.get("bearer", None)
        true_pipeline_name = start_upload_dict.get("pipeline_name")

        if upload_token is None:
            raise ValueError("No upload token found")

        # Login to upload registry
        try:
            docker_client.login(
                username="pipeline",
                password=upload_token,
                registry="http://" + upload_registry,
            )
        except Exception as e:
            _print(f"Failed to login to registry: {e}", "ERROR")
            raise

        _print(f"Successfully logged in to registry {upload_registry}")

        # Override the tag with the pipeline name from catalyst
        image_to_push = true_pipeline_name + ":" + hash_tag
        image_to_push_reg = upload_registry + "/" + image_to_push

    _print(f"Pushing image to upload registry {upload_registry}", "INFO")

    docker_client.images.get(pipeline_name).tag(image_to_push_reg)
    # Do this after tagging, because we need to use
    # the old pipeline name to tag the local image
    if true_pipeline_name:
        pipeline_name = true_pipeline_name

    _push_docker_image(
        docker_client=docker_client, image=image_to_push_reg, upload_token=upload_token
    )

    new_deployment_request = http.post(
        endpoint="/v4/pipelines",
        json_data=pipelines_schemas.PipelineCreate(
            name=pipeline_name,
            image=image_to_push_reg,
            input_variables=[],
            output_variables=[],
            accelerators=pipeline_config.accelerators,
            description=pipeline_config.description,
            readme=pipeline_config.readme,
            extras=pipeline_config.extras,
            cluster=pipeline_config.cluster,
            scaling_config=pipeline_config.scaling_config_name,
        ).dict(),
    )

    new_deployment = pipelines_schemas.PipelineGet.parse_obj(
        new_deployment_request.json()
    )

    _print(
        f"Created new pipeline deployment for {new_deployment.name} -> {new_deployment.id} (image={new_deployment.image})",  # noqa
        "SUCCESS",
    )

    if pointers:
        pointer_overwrite = getattr(namespace, "pointer_overwrite", False)
        for pointer in pointers:
            create_pointer(pointer, new_deployment.id, force=pointer_overwrite)


def _push_docker_image(
    docker_client: docker.DockerClient, image: str, upload_token: str
):
    resp = docker_client.images.push(
        image,
        stream=True,
        decode=True,
        auth_config=(
            dict(username="pipeline", password=upload_token) if upload_token else None
        ),
    )

    all_ids = []

    current_index = 0

    for line in resp:
        if "error" in line:
            if line["error"] == "unauthorized: authentication required":
                print(
                    """
Failed to authenticate with the registry.
This can happen if your pipeline took longer than an hour to push.
Please try reduce the size of your pipeline or contact mystic.ai"""
                )
                return
            raise ValueError(line["error"])
        elif "status" in line:
            if "id" not in line or line["status"] != "Pushing":
                continue

            if "id" in line and line["id"] not in all_ids:
                all_ids.append(line["id"])
                print("Adding")

            index_difference = all_ids.index(line["id"]) - current_index

            print_string = (
                line["id"] + " " + line["progress"].replace("\n", "").replace("\r", "")
            )

            if index_difference > 0:
                print_string += "\n" * index_difference + "\r"
                # print("up")
            elif index_difference < 0:
                print_string += "\033[A" * abs(index_difference) + "\r"
                # print("down")
            else:
                print_string += "\r"
                # print("same")
            current_index = all_ids.index(line["id"])

            sys.stdout.write(print_string)
            sys.stdout.flush()

    _print(f"Successfully pushed image '{image}' to registry")



from pathlib import Path


def _read_template(template_name: str) -> str:
    template_path = Path(__file__).parent / template_name

    if not template_path.exists():
        raise FileNotFoundError(f"Template {template_name} not found")

    return template_path.read_text()


dockerfile_template = _read_template("dockerfile_template.txt")
pipeline_template_python = _read_template("pipeline_template.py")
readme_template = _read_template("readme_template.md")
turbo_registry_template = _read_template("turbo_registry_template.txt")



import typing as t
from datetime import datetime
from enum import Enum

from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.schemas import BaseModel, pagination
from pipeline.cloud.schemas.cluster import PipelineClusterConfig, PipelineClusterGetLean
from pipeline.cloud.schemas.runs import RunIOType


class IOVariable(BaseModel):
    run_io_type: RunIOType

    title: str | None
    description: str | None
    examples: list[t.Any] | None
    gt: int | None
    ge: int | None
    lt: int | None
    le: int | None
    multiple_of: int | None
    allow_inf_nan: bool | None
    max_digits: int | None
    decimal_places: int | None
    min_length: int | None
    max_length: int | None
    choices: list[t.Any] | None
    dict_schema: t.List[dict] | None
    default: t.Any | None
    optional: bool | None


class PipelineStartUpload(BaseModel):
    pipeline_name: str
    pipeline_tag: t.Optional[str]
    cluster: PipelineClusterConfig | None = None


class PipelineStartUploadResponse(BaseModel):
    bearer: str
    upload_registry: t.Optional[str]
    #: Full username/pipeline_name. Used for naming docker image
    pipeline_name: str


class PipelineCreate(BaseModel):
    name: str
    image: str

    input_variables: t.List[IOVariable]
    output_variables: t.List[IOVariable]

    accelerators: t.Optional[t.List[Accelerator]]

    cluster: PipelineClusterConfig | None = None

    scaling_config: str | None = None

    # Additional meta data
    description: t.Optional[str]
    readme: t.Optional[str]
    extras: t.Optional[dict]


class Pipeline(BaseModel):
    name: str
    image: str

    input_variables: t.List[IOVariable]
    output_variables: t.List[IOVariable]
    extras: t.Optional[dict]


class PipelinePatch(BaseModel):
    input_variables: t.Optional[t.List[IOVariable]]
    output_variables: t.Optional[t.List[IOVariable]]

    accelerators: t.Optional[t.List[Accelerator]]

    extras: t.Optional[dict]
    #: The name of the scaling configuration
    scaling_config: str | None = None


class PipelineListPagination(pagination.Pagination):
    class OrderBy(str, Enum):
        created_at = "created_at"
        updated_at = "updated_at"

        pipeline_name = "pipeline_name"
        image = "image"

    order_by: OrderBy
    order: pagination.Order


class PipelineDeploymentStatus(str, Enum):
    not_deployed = "not_deployed"
    deploying = "deploying"
    deployed = "deployed"
    failed = "failed"
    deleting = "deleting"
    deleted = "deleted"


class PipelineState(str, Enum):
    not_loaded = "not_loaded"
    loading = "loading"
    loaded = "loaded"
    load_failed = "load_failed"
    startup_failed = "startup_failed"
    container_failed = "container_failed"

    # backwards compatability
    failed = "failed"


class PipelineContainerState(BaseModel):
    state: PipelineState
    message: t.Optional[str]


class PipelineScalingInfo(BaseModel):
    current_replicas: int
    desired_replicas: int
    current_pipeline_states: dict[PipelineState, int]


class PipelineGet(Pipeline):
    id: str

    created_at: datetime
    updated_at: datetime

    accelerators: t.Optional[t.List[Accelerator]]

    cluster: PipelineClusterGetLean | None = None

    extras: t.Optional[dict]
    #: The name of the scaling configuration
    scaling_config: str | None = None
    #: Additional info attached when pipeline fails to load or startup
    failed_state_info: t.Optional[PipelineContainerState]



from argparse import ArgumentParser, Namespace, _SubParsersAction

from pipeline import current_configuration
from pipeline.cloud import authenticate
from pipeline.util.logging import _print


def login_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    login_parser = command_parser.add_parser(
        "login", help="Login with a compute cluster."
    )
    login_parser.set_defaults(func=_login)

    login_parser.add_argument(
        "-u",
        "--url",
        type=str,
        required=False,
        help="Remote URL for auth (default URL is https://www.mystic.ai)",
        default="https://www.mystic.ai",
    )

    login_parser.add_argument(
        "alias",
        help="Name to use for cluster.",
    )

    login_parser.add_argument(
        "token",
        help="API token for cluster",
    )

    login_parser.add_argument(
        "-a",
        "--active",
        help="Set as the current active remote",
        action="store_true",
    )


def use_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    use_parser = command_parser.add_parser(
        "use",
        help="Use a compute cluster.",
    )
    use_parser.set_defaults(func=_use)

    use_parser.add_argument(
        "alias",
        help="Name of the cluster to use.",
    )


def remove_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    ...


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "get",
        help="Get information about the current remotes.",
    )
    get_parser.set_defaults(func=_get)


def _login(namespace: Namespace) -> None:
    alias = getattr(namespace, "alias")
    url = getattr(namespace, "url")
    token = getattr(namespace, "token")
    active = getattr(namespace, "active", False)

    if current_configuration.remotes and any(
        [remote.alias == alias for remote in current_configuration.remotes]
    ):
        _print(f"Alias '{alias}' already exists", level="ERROR")
        return

    try:
        authenticate(
            token,
            url,
        )
    except Exception:
        _print(f"Couldn't authenticate with {url}", level="ERROR")
        return
    current_configuration.add_remote(
        alias=alias,
        url=url,
        token=token,
    )

    if active:
        _print(f"Setting new remote as active '{alias}'")
        current_configuration.set_active_remote(alias)

    _print(f"Successfully authenticated with {alias}")


def _use(namespace: Namespace) -> None:
    alias = getattr(namespace, "alias")
    if current_configuration.remotes is None:
        _print("No remote configurations. Login first.", level="ERROR")
        return

    if not any([remote.alias == alias for remote in current_configuration.remotes]):
        _print(f"Remote '{alias}' not found", level="ERROR")
        return

    current_configuration.set_active_remote(alias)
    _print(f"Set new default remote to '{alias}'")


def _get(namespace: Namespace) -> None:
    if current_configuration.remotes is None:
        _print("No remote configurations. Login first.", level="ERROR")
        return

    remotes = [
        f"{_remote} (active)"
        if _remote is current_configuration.active_remote
        else f"{_remote}"
        for _remote in current_configuration.remotes
    ]

    _print("Authenticated remotes:")
    [print(_remote) for _remote in remotes]



from pydantic import BaseModel


class PipelineClusterConfig(BaseModel):
    id: str
    node_pool: str


class PipelineClusterGetLean(BaseModel):
    """A lean representation of a cluster when returned from an API call"""

    id: str
    node_pool: str
    # Optional for backward compatibility
    name: str | None = None
    provider: str | None = None



from argparse import Namespace
from pathlib import Path

import docker
import docker.errors
import yaml

from pipeline.container import docker_templates
from pipeline.util.logging import _print

from .schemas import PipelineConfig


def build_container(namespace: Namespace):
    _print("Starting build service...", "INFO")
    template = docker_templates.dockerfile_template

    config_file = Path(getattr(namespace, "file", "./pipeline.yaml"))

    if not config_file.exists():
        raise FileNotFoundError(f"Config file {config_file} not found")

    config = config_file.read_text()
    pipeline_config_yaml = yaml.load(config, Loader=yaml.FullLoader)

    pipeline_config = PipelineConfig.parse_obj(pipeline_config_yaml)

    if pipeline_config.extras and pipeline_config.extras.get("turbo_registry", False):
        template = docker_templates.turbo_registry_template

    if not pipeline_config.runtime:
        raise ValueError("No runtime config found")
    if not pipeline_config.runtime.python:
        raise ValueError("No python runtime config found")

    python_runtime = pipeline_config.runtime.python
    dockerfile_path = getattr(namespace, "docker_file", None)
    if dockerfile_path is None:
        dockerfile_str = template.format(
            python_version=python_runtime.version,
            python_requirements=(
                " ".join(python_runtime.requirements)
                if python_runtime.requirements
                else ""
            ),
            container_commands="".join(
                [
                    "RUN " + command + " \n"
                    for command in pipeline_config.runtime.container_commands or []
                ]
            ),
            pipeline_path=pipeline_config.pipeline_graph,
            pipeline_name=pipeline_config.pipeline_name,
            pipeline_image=pipeline_config.pipeline_name,
        )

        dockerfile_path = Path("./pipeline.dockerfile")
        dockerfile_path.write_text(dockerfile_str)
    else:
        dockerfile_path = Path(dockerfile_path)
    docker_client = docker.APIClient()
    generator = docker_client.build(
        path="./",
        dockerfile=dockerfile_path.absolute(),
        rm=True,
        decode=True,
        platform="linux/amd64",
    )
    docker_image_id = None
    while True:
        try:
            output = generator.__next__()
            if "aux" in output:
                docker_image_id = output["aux"]["ID"]
            if "stream" in output:
                _print(output["stream"].strip("\n"))
            if "errorDetail" in output:
                raise Exception(output["errorDetail"])
        except StopIteration:
            _print("Docker image build complete.")
            break

    docker_client = docker.from_env()
    new_container = docker_client.images.get(docker_image_id)

    created_image_full_id = new_container.id.split(":")[1]
    created_image_short_id = created_image_full_id[:12]

    _print(f"Built container {created_image_short_id}", "SUCCESS")

    pipeline_repo = (
        pipeline_config.pipeline_name.split(":")[0]
        if ":" in pipeline_config.pipeline_name
        else pipeline_config.pipeline_name
    )

    new_container.tag(pipeline_repo)
    _print(f"Created tag {pipeline_repo}", "SUCCESS")

    new_container.tag(pipeline_repo, tag=created_image_short_id)
    _print(f"Created tag {pipeline_repo}:{created_image_short_id}", "SUCCESS")



from pipeline import Pipeline, Variable, entity, pipe


# Put your model inside of the below entity class
@entity
class MyModelClass:
    @pipe(run_once=True, on_startup=True)
    def load(self) -> None:
        # Perform any operations needed to load your model here
        print("Loading model...")

        ...

        print("Model loaded!")

    @pipe
    def predict(self, output_number: int) -> str:
        # Perform any operations needed to predict with your model here
        print("Predicting...")

        ...

        print("Prediction complete!")

        return f"Your number: {output_number}"


with Pipeline() as builder:
    input_var = Variable(
        int,
        description="A basic input number to do things with",
        title="Input number",
    )

    my_model = MyModelClass()
    my_model.load()

    output_var = my_model.predict(input_var)

    builder.output(output_var)

my_new_pipeline = builder.get_pipeline()



import typing as t

from pydantic import BaseModel

from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.schemas import cluster as cluster_schemas


class PythonRuntime(BaseModel):
    version: str
    requirements: list[str] | None

    class Config:
        extra = "forbid"


class RuntimeConfig(BaseModel):
    container_commands: list[str] | None
    python: PythonRuntime | None

    class Config:
        extra = "forbid"


class PipelineConfig(BaseModel):
    runtime: RuntimeConfig
    accelerators: list[Accelerator] = []
    accelerator_memory: int | None
    pipeline_graph: str
    pipeline_name: str = ""
    description: str | None = None
    readme: str | None = None
    extras: dict[str, t.Any] | None
    cluster: cluster_schemas.PipelineClusterConfig | None = None
    scaling_config_name: str | None = None

    class Config:
        extra = "forbid"



class RunInputException(Exception):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class RunnableError(Exception):
    """Exception raised during the execution of a pipeline."""

    def __init__(
        self,
        exception: Exception,
        traceback: str | None,
    ) -> None:
        self.exception = exception
        self.traceback = traceback
        super().__init__(self.exception, self.traceback)

    def __str__(self):
        return f"RunnableError({repr(self.exception)})"



import importlib
import importlib.metadata
import io
import random
import string
from typing import Any, Callable, Optional


def package_version() -> str:
    """Return the version of the installed `pipeline-ai` package."""
    return importlib.metadata.version("pipeline-ai")


def generate_id(length: int) -> str:
    return "".join((random.choice(string.ascii_letters) for i in range(length)))


def python_object_to_name(obj: Any) -> Optional[str]:
    # Consider limiting the size of the name in future releases
    name = getattr(obj, "__name__", str(obj))

    return name


# def python_object_to_file_create(obj: Any, name: str = None):
#     if name is None:
#         name = generate_id(20)

#     return FileCreate(name=name, file_bytes=python_object_to_hex)


class CallbackBytesIO(io.BytesIO):
    """Provides same interface as BytesIO but additionally calls a callback function
    whenever the 'read' method is called.

    This is similar to tqdm's own CallbackIOWrapper but this does not play nicely with
    all features of httpx so we use our own in some cases.
    """

    def __init__(self, callback: Callable, initial_bytes: bytes):
        self._callback = callback
        super().__init__(initial_bytes)

    def read(self, size=-1) -> bytes:
        data = super().read(size)
        self._callback(len(data))
        return data



import typing
from functools import partial, wraps

from pipeline.objects.function import Function
from pipeline.objects.graph import GraphNode, Variable
from pipeline.objects.model import Model
from pipeline.objects.pipeline import Pipeline


def pipe(function=None, *, run_once=False, on_startup=False):
    """_summary_

    Args:
        function (callable, optional): _description_. This is the function to be
            wrapped, you do not pass this in manually it's automatically handled.

        run_once (bool, optional): _description_. Defaults to False. Setting to True
            will ensure that the decorated funciton is only called once
            in a pipeline run

        on_startup (bool, optional): _description_. Defaults to False. Setting to True
            will cause the wrapped function to be executed at the start of a pipeline
            run, regardless of when it's placed when defining the pipeline.

    """
    if function is None:
        return partial(pipe, run_once=run_once, on_startup=on_startup)

    @wraps(function)
    def execute_func(*args, **kwargs):
        if not Pipeline._pipeline_context_active:
            return function(*args, **kwargs)
        else:
            if "return" not in function.__annotations__:
                raise Exception(
                    (
                        "You must define an output type for a pipeline function. "
                        "e.g. def my_func(...) -> float:"
                    )
                )

            processed_args: Variable = []
            for input_arg in args:
                if isinstance(input_arg, Variable):
                    processed_args.append(input_arg)
                elif hasattr(input_arg, "__entity__"):
                    if function.__pipe__.class_instance is None:
                        function.__pipe__.class_instance = input_arg
                elif isinstance(input_arg, tuple) and all(
                    isinstance(var, Variable) for var in input_arg
                ):
                    raise Exception(
                        "Must seperate outputs from functions with Tuple outputs:"
                        "`var1, var2, ..., varN = func(...)`",
                    )

                else:
                    print(hasattr(function, "__self__"))
                    print(f"Function: {function}, Input: {input_arg}")
                    # breakpoint()

                    if function in input_arg.__dict__.values():
                        print("Function found in class")
                    else:
                        print(input_arg.__dict__.values())

                    raise Exception(
                        (
                            f"Can only input pipeline variables to a function "
                            f"when defining a graph, got: {type(input_arg)}"
                        )
                    )

            node_outputs: typing.List[Variable] = []

            function_output = function.__annotations__["return"]
            if getattr(function_output, "__origin__", None) == tuple:
                context_manager_variables = node_outputs = tuple(
                    Variable(type_class=output_variable, is_input=False)
                    for output_variable in function_output.__args__
                )

            else:
                context_manager_variables = Variable(
                    type_class=function.__annotations__["return"],
                    is_input=False,
                )
                node_outputs = [context_manager_variables]

            # Pipeline.add_variables(*node_outputs)
            Pipeline.add_function(function.__pipe__)

            new_node = GraphNode(
                function=function.__pipe__,
                inputs=processed_args,
                outputs=node_outputs,
            )

            Pipeline.add_graph_node(new_node)

            return context_manager_variables

    execute_func.__function__ = function

    function.__run_once__ = run_once
    function.__has_run__ = False

    function.__on_startup__ = on_startup
    function.__pipe__ = Function(function)

    return execute_func


class entity(object):
    def __init__(
        self,
        model_class=None,
    ):
        if model_class is not None:
            model_class.__entity__ = True

        self.model_class = model_class

    def __call__(self, *args, **kwargs):
        if len(args) + len(kwargs) == 1:
            self.model_class = args[0]
            self.model_class.__entity__ = True
            return self.__function_exe__
        else:
            return self.__function_exe__(*args, **kwargs)

    def __function_exe__(self, *args, **kwargs):
        if not Pipeline._current_pipeline:
            return self.model_class(*args, **kwargs)
        else:
            created_model = self.model_class(*args, **kwargs)

            model_schema = Model(model=created_model)
            Pipeline._current_pipeline.models.append(model_schema)
            return created_model
